/*
 * Copyright (c) 2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Mar 29 2017
 */
#include "acheader.h"

#include <stdint.h>

#include "arduio.h"


/* --- local data ---------------------------------------------------------- */

static inline void	bitbang0(volatile uint8_t *port, uint8_t bit);
static inline void	bitbang1(volatile uint8_t *port, uint8_t bit);


/* --- ws28xx_set ---------------------------------------------------------- */

void
ws28xx_set(volatile uint8_t *port, uint8_t bit, uint16_t v)
{
  uint8_t i, c;

  for(c = 0; c < 3; c++) {
    for(i = 8; i > 0; i--) {
      if (i > 4 && v & 0x800)
        bitbang1(port, bit);
      else
        bitbang0(port, bit);

      if (i > 4) v <<= 1;
      ario_poll_uart();
    }
  }
}


/* --- bitbang ------------------------------------------------------------- */

static inline __attribute__((always_inline)) void
bitbang0(volatile uint8_t *port, uint8_t bit)
{
  uint8_t tmp[2];

  __asm__ volatile(
    "ld %[tmp], %a[port]"	"\n\t"
    "or %[tmp], %[bit]"		"\n\t"
    "st %a[port], %[tmp]"	"\n\t" /* set bit high */
    "nop"			"\n\t" /* delay by 4+2 (375ns) */
    "nop"			"\n\t"
    "com %[bit]"		"\n\t"
    "and %[tmp], %[bit]"	"\n\t"
    "st %a[port], %[tmp]"	"\n\t" /* set bit low */
    : [bit] "=r" (tmp[0]), [tmp] "=r" (tmp[1])
    : [port] "e" (port), "[bit]" (bit));
}

static inline  __attribute__((always_inline)) void
bitbang1(volatile uint8_t *port, uint8_t bit)
{
  uint8_t tmp[3];

  __asm__ volatile(
    "ld %[tmp], %a[port]"	"\n\t"
    "or %[tmp], %[bit]"		"\n\t"
    "st %a[port], %[tmp]"	"\n\t" /* set bit high */
    "ldi %[i], 3"		"\n\t" /* delay by 1+(3*3-1)+3+2 (875ns) */
    "1: dec %[i]"		"\n\t"
    "brne 1b"			"\n\t"
    "nop"			"\n\t"
    "com %[bit]"		"\n\t"
    "and %[tmp], %[bit]"	"\n\t"
    "st %a[port], %[tmp]"	"\n\t" /* set bit low */
    : [bit] "=r" (tmp[0]), [tmp] "=r" (tmp[1]), [i] "=r" (tmp[2])
    : [port] "e" (port), "[bit]" (bit));
}
